package com.lzmispt.data.resource.service.datasync.impl.flink;

package .tmp;

import com.alibaba.druid.pool.DruidPooledConnection;
import com.alibaba.fastjson.JSON;
import com.lzmispt.core.dao.database.impl.BaseDaoImpl;
import com.lzmispt.core.entity.coreconst.CoreConst;
import com.lzmispt.core.entity.dto.authorize.UserInfoDTO;
import com.lzmispt.core.service.encrypt.EncryptManagerFactory;
import com.lzmispt.data.center.entity.dto.databasestruct.ColumnParamDTO;
import com.lzmispt.data.center.entity.dto.databasestruct.DataSourceDTO;
import com.lzmispt.data.center.entity.dto.databasestruct.TableParamDTO;
import com.lzmispt.data.center.entity.enums.BaseSysFieldEnum;
import com.lzmispt.data.center.entity.enums.StatusCodeEnum;
import com.lzmispt.data.center.entity.vo.resource.datasource.DataSourceVO;
import com.lzmispt.data.center.entity.vo.resource.metadata.MetadataEntityVO;
import com.lzmispt.data.center.entity.vo.standard.metadata.MetadataEntityDataItemVO;
import com.lzmispt.data.center.entity.vo.standard.metadata.MetadataFieldWithElementVO;
import com.lzmispt.data.center.entity.vo.standard.metadata.MetadataTypeVO;
import com.lzmispt.data.resource.client.DataStandardClient;
import com.lzmispt.data.resource.dao.datasync.IDataSyncDao;
import com.lzmispt.data.resource.dao.metadata.IMetadataDao;
import com.lzmispt.data.resource.entity.dto.datasync.DataSyncDataLogQueryDTO;
import com.lzmispt.data.resource.entity.dto.datasync.FlinkTaskInfoDTO;
import com.lzmispt.data.resource.entity.dto.resanddata.ResAndDataEntityDTO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncLogPO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncTaskFieldsPO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncTaskPO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncTaskTablesPO;
import com.lzmispt.data.resource.entity.po.metadata.MetadataEntityInfoItemPO;
import com.lzmispt.data.resource.entity.po.metadata.MetadataEntityPO;
import com.lzmispt.data.resource.entity.vo.datasync.DataSyncTaskLogVO;
import com.lzmispt.data.resource.entity.vo.datasync.ReadWriteStatistics;
import com.lzmispt.data.resource.entity.vo.datasync.TableColumnsVO;
import com.lzmispt.data.resource.entity.vo.datasync.TaskReadWriteStatistics;
import com.lzmispt.data.resource.manager.metadata.MetadataManager;
import com.lzmispt.data.resource.service.datasync.impl.DataSyncThreadPool;
import com.lzmispt.data.resource.service.datasync.impl.flink.FlinkApi;
import com.lzmispt.data.resource.service.datasync.impl.handle.FieldTypeValidatorHandler;
import com.lzmispt.data.resource.service.datasync.impl.statistics.DataSyncStatisticsHandler;
import com.lzmispt.data.resource.service.metadata.IMetadataField;
import com.lzmispt.dcbase.ServiceUtil;
import com.lzmispt.dcbase.databasestruct.DataStructProvider;
import com.lzmispt.dcbase.dcconst.DCConst;
import com.lzmispt.dcbase.sysfield.SysFieldHelper;
import com.lzmispt.dcbase.util.TokenUtil;
import com.lzmispt.flink.entity.cdc.ColumnStruct;
import com.lzmispt.flink.entity.config.ConnectInfo;
import com.lzmispt.flink.entity.config.FieldMapping;
import com.lzmispt.flink.entity.config.SyncConfig;
import com.lzmispt.flink.entity.config.SyncSetting;
import com.lzmispt.flink.entity.config.TableInfo;
import com.lzmispt.flink.entity.config.TableMapping;
import com.lzmispt.flink.statistics.StatisticsRecorder;
import com.lzmispt.flink.util.DataSourceHolder;
import com.lzmispt.flink.util.DatabaseHelper;
import com.lzmispt.framework.entity.entity.ResultEntity;
import com.lzmispt.framework.exception.BusinessException;
import com.lzmispt.framework.exception.ExceptionUtil;
import com.lzmispt.framework.log.LogHelper;
import com.lzmispt.framework.utils.WorkerId;
import com.ververica.cdc.debezium.utils.datacenter.DataSyncLogMsg;
import java.sql.PreparedStatement;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;
import org.apache.flink.runtime.jobgraph.SavepointRestoreSettings;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.domain.Sort;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.CriteriaDefinition;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.data.mongodb.core.query.UpdateDefinition;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

@Service
public class DataSyncExecutor {
  @Autowired
  @Qualifier("leadingDataLogMongo")
  MongoTemplate mongoTemplate;
  
  @Autowired
  DataSyncStatisticsHandler statisticsHandler;
  
  private static final DateTimeFormatter TIME_FORMAT = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
  
  @Autowired
  private BaseDaoImpl baseDao;
  
  @Autowired
  private IDataSyncDao dataSyncDao;
  
  @Autowired
  private DataStandardClient standardClient;
  
  @Autowired
  private IMetadataField metadataField;
  
  private static final String LEVEL_INFO = "INFO";
  
  private static final String DEFAULT_TOKEN = "0";
  
  public void startDataSync(Long taskId, String tokenId, boolean async) throws BusinessException {
    DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
    if (task == null)
      return; 
    startDataSync(task, tokenId, async);
  }
  
  public void startDataSync(DataSyncTaskPO task, String tokenId, boolean async) throws BusinessException {
    Integer orignalStatus = task.getRunStatus();
    if (DCConst.DataSyncTaskRunStatus.IN_PROGRESS.equals(orignalStatus))
      return; 
    TaskReadWriteStatistics statistics = (TaskReadWriteStatistics)JSON.parseObject(task.getReadWriteStatistics(), TaskReadWriteStatistics.class);
    if (statistics.getErrorRecordCount() >= task.getDirtyDataThresholdValue().intValue())
      throw new BusinessException(Integer.valueOf(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_DIRTY_DATA_EXCEED_THRESHOLD.getCode()), 
          String.format(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_DIRTY_DATA_EXCEED_THRESHOLD.getMessage(), new Object[] { task.getDirtyDataThresholdValue() })); 
    List<DataSyncTaskTablesPO> mappings = this.dataSyncDao.selectSyncTaskDataBaseNameAndDataTableByTaskId(task.getId());
    SyncConfig taskConfig = buildTaskConfig(task, tokenId);
    Map<Long, SyncConfig> syncConfigMap = replenishConfigAndCheckTable(mappings, taskConfig);
    DataSyncTaskPO update = new DataSyncTaskPO();
    update.setId(task.getId());
    update.setUpdateTime(new Date());
    update.setRunStatus(DCConst.DataSyncTaskRunStatus.IN_PROGRESS);
    if (task.getFirstRunTime() == null)
      update.setFirstRunTime(new Date()); 
    this.baseDao.updateSelective(update);
    if (DCConst.DataSyncTaskRunStatus.PAUSE_OPERATION.equals(orignalStatus) || DCConst.DataSyncTaskRunStatus.TERMINATION.equals(orignalStatus)) {
      UserInfoDTO userInfo = Optional.<UserInfoDTO>ofNullable(TokenUtil.getUserInfoByTokenId(tokenId)).orElse(new UserInfoDTO());
      recordSyncLog(task.getId(), new DataSyncLogMsg(null, "任务已恢复", "INFO", TIME_FORMAT.format(LocalDateTime.now())), userInfo.getUserId(), userInfo.getOrganizationId());
    } 
    if (async) {
      batchStartFlinkJob(mappings, syncConfigMap);
    } else {
      DataSyncThreadPool.concurrent().submit(() -> batchStartFlinkJob(mappings, syncConfigMap));
    } 
  }
  
  private void batchStartFlinkJob(List<DataSyncTaskTablesPO> mappings, Map<Long, SyncConfig> syncConfigMap) {
    if (mappings.isEmpty())
      LogHelper.info("没有要执行的表映射任务，跳过处理"); 
    for (DataSyncTaskTablesPO mapping : mappings) {
      SyncConfig syncConfig = syncConfigMap.get(mapping.getId());
      FlinkTaskInfoDTO taskInfo = (FlinkTaskInfoDTO)JSON.parseObject(mapping.getDataSyncTaskInfo(), FlinkTaskInfoDTO.class);
      if (taskInfo != null && FlinkApi.isJobRunning(taskInfo.getCurrent_job_id())) {
        LogHelper.info("jobId：" + taskInfo.getCurrent_job_id() + "，作业正在执行中，不再重复执行");
        continue;
      } 
      SavepointRestoreSettings savepoint = null;
      if (taskInfo == null || !DCConst.DataSyncTaskStage.INCREMENT.equals(taskInfo.getCurrent_sync_stage())) {
        taskInfo = clearTableBeforeSync(mapping, taskInfo, syncConfig);
      } else {
        savepoint = FlinkApi.getSavepoint(taskInfo.getCurrent_job_id());
        if (savepoint == null)
          taskInfo = clearTableBeforeSync(mapping, taskInfo, syncConfig); 
      } 
      startFlinkJob(syncConfig, taskInfo, savepoint);
    } 
  }
  
  public void stopDataSync(Long taskId, String tokenId) {
    DataSyncTaskPO load = this.dataSyncDao.selectDataSyncTaskById(taskId);
    if (load == null)
      return; 
    DataSourceVO dataSource = (DataSourceVO)this.standardClient.getPlaintextDataSourceById(load.getSourceDataSourceId(), "Collection", tokenId).getObj();
    String dbType = dataSource.getType();
    DataSyncTaskPO taskUpdate = new DataSyncTaskPO();
    taskUpdate.setId(taskId);
    taskUpdate.setRunStatus(DCConst.DataSyncTaskRunStatus.PAUSE_OPERATION);
    taskUpdate.setUpdateTime(new Date());
    this.baseDao.updateSelective(taskUpdate);
    List<DataSyncTaskTablesPO> mappings = this.dataSyncDao.selectSyncTaskDataBaseNameAndDataTableByTaskId(taskId);
    mappings.forEach(item -> stopFlinkJob((FlinkTaskInfoDTO)JSON.parseObject(item.getDataSyncTaskInfo(), FlinkTaskInfoDTO.class), dbType));
    UserInfoDTO userInfo = Optional.<UserInfoDTO>ofNullable(TokenUtil.getUserInfoByTokenId(tokenId)).orElse(new UserInfoDTO());
    recordSyncLog(taskId, new DataSyncLogMsg(null, "任务已暂停", "INFO", TIME_FORMAT.format(LocalDateTime.now())), userInfo.getUserId(), userInfo.getOrganizationId());
  }
  
  public void recoverDataSync(Long taskId, String tokenId) throws BusinessException {
    startDataSync(taskId, tokenId, false);
  }
  
  public void rerunDirtyData(Long taskId, List<Long> tableIds, String tokenId) throws BusinessException {
    if (CollectionUtils.isEmpty(tableIds))
      return; 
    UserInfoDTO userInfo = Optional.<UserInfoDTO>ofNullable(TokenUtil.getUserInfoByTokenId(tokenId)).orElse(new UserInfoDTO());
    DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
    if (DCConst.DataSyncTaskRunStatus.PAUSE_OPERATION.equals(task.getRunStatus())) {
      recordSyncLog(task.getId(), new DataSyncLogMsg(null, "任务暂未运行，请启动任务后重新执行脏数据重跑", "INFO", TIME_FORMAT.format(LocalDateTime.now())), userInfo.getUserId(), userInfo.getOrganizationId());
      throw new BusinessException(Integer.valueOf(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_TASK_IN_PAUSE_CAN_NOT_RERUN_DIRTY_DATA.getCode()), StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_TASK_IN_PAUSE_CAN_NOT_RERUN_DIRTY_DATA.getMessage());
    } 
    List<DataSyncTaskTablesPO> mappings = this.dataSyncDao.selectDataSyncTableByTableIds(tableIds);
    SyncConfig taskConfig = buildTaskConfig(task, tokenId);
    String dbType = taskConfig.getSourceConnection().getType();
    Map<Long, SyncConfig> syncConfigMap = replenishConfigAndCheckTable(mappings, taskConfig);
    Map<Long, Integer> sourceDataCountMap = new HashMap<>();
    syncConfigMap.forEach((key, value) -> {
          DataStructProvider obj = (DataStructProvider)DataStructProvider.getDataBaseProvider(parseToDataSourceDTO(value.getSourceConnection()), false, false).getObj();
          sourceDataCountMap.put(key, Integer.valueOf(obj.getTableDataCount(value.getMapping().getSource().getTableName())));
        });
    ReadWriteStatistics totalSub = new ReadWriteStatistics();
    mappings.forEach(item -> {
          TableInfo target = ((SyncConfig)syncConfigMap.get(item.getId())).getMapping().getTarget();
          String fullTableName = DatabaseHelper.getFullTableName(target.getDatabase(), target.getSchema(), target.getTableName());
          recordSyncLog(task.getId(), new DataSyncLogMsg(null, "开始重跑目标表[" + fullTableName + "]脏数据", "INFO", TIME_FORMAT.format(LocalDateTime.now())), userInfo.getUserId(), userInfo.getOrganizationId());
          ReadWriteStatistics statistics = (ReadWriteStatistics)JSON.parseObject(item.getReadWriteStatistics(), ReadWriteStatistics.class);
          DataSyncStatisticsHandler.plus(totalSub, statistics);
          DataSyncTaskTablesPO update = new DataSyncTaskTablesPO();
          update.setId(item.getId());
          update.setReadWriteStatistics(DataSyncStatisticsHandler.TABLE_STATISTICS_INIT);
          this.baseDao.updateSelective(update);
          clearDirtyDataRecord(item.getId());
        });
    this.statisticsHandler.subTaskStatistics(taskId, totalSub);
    for (DataSyncTaskTablesPO mapping : mappings) {
      FlinkTaskInfoDTO flinkInfo = (FlinkTaskInfoDTO)JSON.parseObject(mapping.getDataSyncTaskInfo(), FlinkTaskInfoDTO.class);
      if (flinkInfo == null)
        continue; 
      stopFlinkJob(flinkInfo, dbType);
      SyncConfig syncConfig = syncConfigMap.get(mapping.getId());
      clearTableData(mapping, syncConfig);
      flinkInfo.setJob_type(DCConst.DataSyncJobType.RERUN_DIRTY_DATA.intValue());
      Integer sourceDataCount = sourceDataCountMap.get(mapping.getId());
      if (sourceDataCount != null && sourceDataCount.intValue() == 0) {
        String completeTime = TIME_FORMAT.format(LocalDateTime.now());
        startFlinkJob(syncConfig, flinkInfo, null, null);
        recordStageChange(new DataSyncLogMsg(StatisticsRecorder.getKey(syncConfig.getId(), syncConfig.getMapping().getId()), null, "RERUN_COMPLETE", completeTime));
      } 
      startFlinkJob(syncConfig, flinkInfo, null, sourceDataCount);
    } 
  }
  
  public DataSyncTaskLogVO getDataSyncTaskLog(DataSyncDataLogQueryDTO param, String tokenId) {
    Query query = new Query();
    Criteria where = Criteria.where("task_id").is(param.getTaskId()).and("is_delete").is(DCConst.FALSE);
    if (param.getStartTime() != null)
      where.and("create_time").gt(new Date(param.getStartTime().longValue())); 
    query.addCriteria((CriteriaDefinition)where);
    Sort sort = Sort.by(new Sort.Order[] { Sort.Order.asc("create_time") });
    query.with(sort);
    List<DataSyncLogPO> logs = this.mongoTemplate.find(query, DataSyncLogPO.class, "dc_data_sync_log");
    DataSyncTaskLogVO result = new DataSyncTaskLogVO();
    if (!logs.isEmpty()) {
      result.setReturnTime(Long.valueOf(((DataSyncLogPO)logs.get(logs.size() - 1)).getCreateTime().getTime()));
    } else {
      result.setReturnTime(param.getStartTime());
    } 
    result.setLogs((List)logs.stream().map(DataSyncLogPO::getMsg).collect(Collectors.toList()));
    return result;
  }
  
  private FlinkTaskInfoDTO clearTableBeforeSync(DataSyncTaskTablesPO mapping, FlinkTaskInfoDTO taskInfo, SyncConfig syncConfig) {
    FlinkTaskInfoDTO initTaskInfo = new FlinkTaskInfoDTO();
    if (taskInfo != null) {
      ReadWriteStatistics statistics = (ReadWriteStatistics)JSON.parseObject(mapping.getReadWriteStatistics(), ReadWriteStatistics.class);
      this.statisticsHandler.subTaskStatistics(mapping.getDataSyncTaskId(), statistics);
      DataSyncTaskTablesPO tableUpdate = new DataSyncTaskTablesPO();
      tableUpdate.setId(mapping.getId());
      tableUpdate.setReadWriteStatistics(DataSyncStatisticsHandler.TABLE_STATISTICS_INIT);
      this.baseDao.updateSelective(tableUpdate);
      clearDirtyDataRecord(mapping.getId());
      initTaskInfo.setCurrent_sync_stage(taskInfo.getCurrent_sync_stage());
    } 
    clearTableData(mapping, syncConfig);
    return initTaskInfo;
  }
  
  private void clearTableData(DataSyncTaskTablesPO mapping, SyncConfig syncConfig) {
    ConnectInfo targetConn = syncConfig.getTargetConnection();
    TableInfo targetTable = syncConfig.getMapping().getTarget();
    String targetTableName = DatabaseHelper.getTableNameSqlStr(targetTable.getSchema(), targetTable.getTableName(), targetConn.getType());
    String deleteSql = "DELETE FROM " + targetTableName;
    try(DruidPooledConnection conn = DataSourceHolder.get(targetConn).getConnection(); 
        PreparedStatement statement = conn.prepareStatement(deleteSql)) {
      statement.execute();
    } catch (Exception e) {
      recordSyncLog(mapping.getDataSyncTaskId(), new DataSyncLogMsg(null, "清空目标表[" + 
            DatabaseHelper.getFullTableName(targetTable.getDatabase(), targetTable.getSchema(), targetTable.getTableName()) + "]异常：" + ExceptionUtil.getStackTraceMsg(e), "ERROR", TIME_FORMAT
            .format(LocalDateTime.now())));
    } 
  }
  
  private void clearDirtyDataRecord(Long mappingId) {
    Update dirtyUpdate = new Update();
    dirtyUpdate.set("is_delete", DCConst.TRUE);
    Query query = new Query();
    Criteria criteria = (new Criteria("data_sync_table_id")).is(mappingId);
    query.addCriteria((CriteriaDefinition)criteria);
    this.mongoTemplate.updateMulti(query, (UpdateDefinition)dirtyUpdate, "dc_data_sync_dirty_data");
  }
  
  public void startFlinkJob(SyncConfig syncConfig, FlinkTaskInfoDTO flinkTaskInfo, SavepointRestoreSettings savepoint) {
    startFlinkJob(syncConfig, flinkTaskInfo, savepoint, null);
  }
  
  public void startFlinkJob(SyncConfig syncConfig, FlinkTaskInfoDTO flinkTaskInfo, SavepointRestoreSettings savepoint, Integer snapshotCount) {
    String jobId;
    Long taskId = Long.valueOf(syncConfig.getId());
    Long mappingId = Long.valueOf(syncConfig.getMapping().getId());
    try {
      jobId = FlinkApi.executeDataSync(syncConfig, savepoint, snapshotCount);
      LogHelper.info("flink作业启动成功，jobId:" + jobId);
    } catch (Exception e) {
      TableInfo sourceTable = syncConfig.getMapping().getSource();
      String fullTableName = DatabaseHelper.getFullTableName(sourceTable.getDatabase(), sourceTable.getSchema(), sourceTable.getTableName());
      LogHelper.error("[" + fullTableName + "]Flink作业启动异常", ExceptionUtil.getStackTraceMsg(e));
      recordSyncLog(taskId, new DataSyncLogMsg(null, "[" + fullTableName + "]Flink作业启动异常，请尝试暂停任务重新启动: " + ExceptionUtil.getStackTraceMsg(e), "ERROR", TIME_FORMAT.format(LocalDateTime.now())));
      return;
    } 
    DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
    String sourceDbType = syncConfig.getSourceConnection().getType();
    if (!DCConst.DataSyncTaskRunStatus.IN_PROGRESS.equals(task.getRunStatus()) && !DCConst.DataSyncJobType.RERUN_DIRTY_DATA.equals(Integer.valueOf(flinkTaskInfo.getJob_type()))) {
      LogHelper.info("taskId:" + task.getId() + "，作业未处于运行中状态，暂停flink作业");
      FlinkApi.stopByDbType(jobId, sourceDbType);
      return;
    } 
    flinkTaskInfo.setStoped_job_id(flinkTaskInfo.getCurrent_job_id());
    flinkTaskInfo.setCurrent_job_id(jobId);
    String flinkTaskInfoStr = JSON.toJSONString(flinkTaskInfo);
    DataSyncTaskTablesPO updateTable = new DataSyncTaskTablesPO();
    updateTable.setId(mappingId);
    updateTable.setUpdateTime(new Date());
    updateTable.setDataSyncTaskInfo(flinkTaskInfoStr);
    this.baseDao.updateSelective(updateTable);
  }
  
  private void stopFlinkJob(FlinkTaskInfoDTO flinkInfo, String dbType) {
    if (flinkInfo == null)
      return; 
    FlinkApi.stopByDbType(flinkInfo.getCurrent_job_id(), dbType);
  }
  
  public void recordSyncLog(DataSyncLogMsg logMsg) {
    String[] split = StatisticsRecorder.reverseKey(logMsg.getMark());
    Long taskId = Long.valueOf(split[0]);
    recordSyncLog(taskId, logMsg);
    if ("FATAL".equals(logMsg.getLevel()))
      this.statisticsHandler.terminateDataSyncTask(taskId); 
  }
  
  public void recordSyncLog(Long taskId, DataSyncLogMsg logMsg) {
    recordSyncLog(taskId, logMsg, null, null);
  }
  
  public void recordSyncLog(Long taskId, DataSyncLogMsg logMsg, Long userId, Long orgId) {
    DataSyncLogPO insert = new DataSyncLogPO();
    insert.setId((Long)WorkerId.getNextId(Long.class));
    insert.setTaskId(taskId);
    insert.setMsg(logMsg.getTime() + " " + logMsg.getLevel() + " " + logMsg.getMsg());
    insert.setLevel(logMsg.getLevel());
    Date now = new Date();
    insert.setUserIdRecord(userId);
    insert.setOrgIdRecord(orgId);
    insert.setCreateTime(now);
    insert.setUpdateTime(now);
    insert.setIsDelete(DCConst.FALSE);
    this.mongoTemplate.insert(insert, "dc_data_sync_log");
  }
  
  public void recordStageChange(DataSyncLogMsg logMsg) {
    String[] split = StatisticsRecorder.reverseKey(logMsg.getMark());
    Long taskId = Long.valueOf(split[0]);
    Long mappingId = Long.valueOf(split[1]);
    int stage = (!"SNAPSHOT".equals(logMsg.getLevel()) ? DCConst.DataSyncTaskStage.INCREMENT : DCConst.DataSyncTaskStage.SNAPSHOT).intValue();
    String stageStr = (stage == DCConst.DataSyncTaskStage.INCREMENT.intValue()) ? "增量" : "存量";
    DataSyncTaskTablesPO mapping = this.dataSyncDao.selectDataSyncTaskTableById(mappingId);
    FlinkTaskInfoDTO flinkTaskInfo = (FlinkTaskInfoDTO)JSON.parseObject(mapping.getDataSyncTaskInfo(), FlinkTaskInfoDTO.class);
    String fullTableName = DatabaseHelper.getFullTableName(mapping.getSourceDatabaseName(), mapping.getSourceDatabaseSchema(), mapping.getSourceTableName());
    if (flinkTaskInfo == null)
      return; 
    boolean intoNextStage = (flinkTaskInfo.getCurrent_sync_stage() == null || stage > flinkTaskInfo.getCurrent_sync_stage().intValue());
    boolean isRerun = !DCConst.DataSyncJobType.DATA_SYNC.equals(Integer.valueOf(flinkTaskInfo.getJob_type()));
    boolean rerunDirtyDataFinished = (isRerun && "RERUN_COMPLETE".equals(logMsg.getLevel()));
    boolean dataSyncIntoNextStage = (!isRerun && intoNextStage);
    if (rerunDirtyDataFinished) {
      String completeTime = TIME_FORMAT.format(LocalDateTime.now());
      MetadataEntityPO entity = ((IMetadataDao)ServiceUtil.getMetadataDao("Collection", IMetadataDao.class)).selectIsMasterMetadataEntityByDirectoryId(mapping.getTargetTableDirectoryId());
      DataSourceVO targetDataSource = (DataSourceVO)this.standardClient.getPlaintextDataSourceById(entity.getDataSourceId(), "Collection", "0").getObj();
      recordSyncLog(taskId, new DataSyncLogMsg(null, "目标表[" + DatabaseHelper.getFullTableName(mapping.getTargetDatabaseName(), "mysql".equals(targetDataSource.getType()) ? null : entity.getDataSchemaName(), entity.getEntityTableName()) + "]脏数据重跑完成", "INFO", completeTime));
      if (intoNextStage) {
        String msg = "[" + fullTableName + "]进入增量同步阶段";
        recordSyncLog(taskId, new DataSyncLogMsg(null, msg, "INFO", completeTime));
      } 
      DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
      if (!DCConst.DataSyncTaskRunStatus.IN_PROGRESS.equals(task.getRunStatus())) {
        DataSourceVO dataSource = (DataSourceVO)this.standardClient.getPlaintextDataSourceById(task.getSourceDataSourceId(), "Collection", "0").getObj();
        String dbType = dataSource.getType();
        stopFlinkJob(flinkTaskInfo, dbType);
      } 
    } else if (dataSyncIntoNextStage) {
      if (flinkTaskInfo.getCurrent_sync_stage() == null && stage == DCConst.DataSyncTaskStage.INCREMENT.intValue()) {
        String str = "[" + fullTableName + "]进入存量同步阶段";
        recordSyncLog(taskId, new DataSyncLogMsg(null, str, "INFO", logMsg.getTime()));
      } 
      String msg = "[" + fullTableName + "]进入" + stageStr + "同步阶段";
      recordSyncLog(taskId, new DataSyncLogMsg(null, msg, "INFO", logMsg.getTime()));
    } 
    if (dataSyncIntoNextStage || rerunDirtyDataFinished) {
      DataSyncTaskTablesPO update = new DataSyncTaskTablesPO();
      update.setId(mapping.getId());
      flinkTaskInfo.setCurrent_sync_stage(Integer.valueOf(stage));
      if (rerunDirtyDataFinished)
        flinkTaskInfo.setJob_type(DCConst.DataSyncJobType.DATA_SYNC.intValue()); 
      update.setUpdateTime(new Date());
      update.setDataSyncTaskInfo(JSON.toJSONString(flinkTaskInfo));
      this.baseDao.updateSelective(update);
    } 
  }
  
  private SyncConfig buildTaskConfig(DataSyncTaskPO task, String tokenId) {
    SyncConfig taskConfig = new SyncConfig();
    taskConfig.setId(String.valueOf(task.getId()));
    taskConfig.setSetting(buildSyncSetting(task));
    DataSourceVO sourceVO = getNotEncryptDataSource(task.getSourceDataSourceId(), "Collection", tokenId);
    if (notSuppotDatabaseType(sourceVO.getType()))
      throw new BusinessException(Integer.valueOf(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_DATASOURCE_NOT_SUPPORT.getCode()), 
          String.format(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_DATASOURCE_NOT_SUPPORT.getMessage(), new Object[] { sourceVO.getName() })); 
    taskConfig.setSourceConnection(buildConn(sourceVO));
    DataSourceVO targetVO = getNotEncryptDataSource(task.getTargetDataSourceId(), "Collection", tokenId);
    if (notSuppotDatabaseType(targetVO.getType()))
      throw new BusinessException(Integer.valueOf(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_DATASOURCE_NOT_SUPPORT.getCode()), 
          String.format(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_DATASOURCE_NOT_SUPPORT.getMessage(), new Object[] { targetVO.getName() })); 
    taskConfig.setTargetConnection(buildConn(targetVO));
    return taskConfig;
  }
  
  private Map<Long, SyncConfig> replenishConfigAndCheckTable(List<DataSyncTaskTablesPO> mappings, SyncConfig taskConfig) {
    Map<Long, SyncConfig> syncConfigMap = new HashMap<>();
    for (DataSyncTaskTablesPO mapping : mappings) {
      SyncConfig syncConfig = replenishSyncConfig(taskConfig, mapping);
      syncConfigMap.put(mapping.getId(), syncConfig);
      checkIsTableExist(syncConfig.getMapping().getSource().getTableName(), syncConfig.getSourceConnection());
      checkIsTableExist(syncConfig.getMapping().getTarget().getTableName(), syncConfig.getTargetConnection());
    } 
    return syncConfigMap;
  }
  
  public ConnectInfo buildConn(DataSourceVO source) {
    ConnectInfo info = new ConnectInfo();
    info.setType(source.getType());
    info.setHost(source.getIpName());
    info.setPort(source.getPort());
    info.setUsername(source.getUser());
    info.setPassword(source.getPassword());
    info.setDatabase(source.getDbName());
    info.setSchema(source.getDbSchema());
    return info;
  }
  
  public SyncSetting buildSyncSetting(DataSyncTaskPO task) {
    SyncSetting syncSetting = new SyncSetting();
    syncSetting.setStartTime(SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.START_TIME));
    syncSetting.setSyncStatus(SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.SYNC_STATUS));
    if (DCConst.TRUE.equals(task.getSyncMarkTimestamp()))
      syncSetting.setLastModifyTime(SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.LAST_MODIFY_TIME)); 
    if (DCConst.TRUE.equals(task.getDeleteMode()))
      syncSetting.setIsDelete(SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.IS_DELETE)); 
    syncSetting.setIncludeSchemaChanges(Boolean.valueOf(DCConst.TRUE.equals(task.getSyncTableStructureChange())));
    return syncSetting;
  }
  
  public TableMapping buildTableMapping(DataSyncTaskTablesPO mapping, SyncSetting setting) {
    TableMapping tableMapping = new TableMapping();
    tableMapping.setId(String.valueOf(mapping.getId()));
    List<DataSyncTaskFieldsPO> syncTaskFieldsList = this.dataSyncDao.selectDataTaskFieldListByTaskTableId(mapping.getId());
    tableMapping.setFieldMappings(buildFieldMappings(syncTaskFieldsList, setting));
    ResAndDataEntityDTO entity = MetadataManager.selectIsMasterMetadataEntityByDirectoryId(mapping.getTargetTableDirectoryId(), DCConst.DataResourceCategory.DATA, mapping.getPlatCode(), "0");
    TableInfo sourceTable = new TableInfo(mapping.getSourceTableName(), mapping.getSourceDatabaseName(), mapping.getSourceDatabaseSchema(), getSourceColumns(syncTaskFieldsList));
    TableInfo targetTable = new TableInfo(entity.getEntityTableName(), mapping.getTargetDatabaseName(), mapping.getSourceDatabaseSchema(), null);
    tableMapping.setSource(sourceTable);
    tableMapping.setTarget(targetTable);
    return tableMapping;
  }
  
  private SyncConfig replenishSyncConfig(SyncConfig taskConfig, DataSyncTaskTablesPO mapping) {
    SyncConfig syncConfig = new SyncConfig();
    BeanUtils.copyProperties(taskConfig, syncConfig);
    TableMapping tableMapping = buildTableMapping(mapping, syncConfig.getSetting());
    syncConfig.setMapping(tableMapping);
    syncConfig.getSourceConnection().setDatabase(tableMapping.getSource().getDatabase());
    if ("mysql".equals(syncConfig.getSourceConnection().getType())) {
      syncConfig.getSourceConnection().setSchema(null);
      tableMapping.getSource().setSchema(null);
    } else {
      syncConfig.getSourceConnection().setSchema(tableMapping.getSource().getSchema());
    } 
    syncConfig.getTargetConnection().setDatabase(tableMapping.getTarget().getDatabase());
    if ("mysql".equals(syncConfig.getTargetConnection().getType())) {
      syncConfig.getTargetConnection().setSchema(null);
      tableMapping.getTarget().setSchema(null);
    } else {
      syncConfig.getTargetConnection().setSchema(tableMapping.getTarget().getSchema());
    } 
    return syncConfig;
  }
  
  public List<FieldMapping> buildFieldMappings(List<DataSyncTaskFieldsPO> syncTaskFieldsList, SyncSetting setting) {
    List<Long> infoItemIds = (List<Long>)syncTaskFieldsList.stream().map(DataSyncTaskFieldsPO::getTargetFieldId).distinct().collect(Collectors.toList());
    List<MetadataFieldWithElementVO> metadataFieldWithElementVOS = (List<MetadataFieldWithElementVO>)this.standardClient.getFieldWithElementByIds(infoItemIds, DCConst.DataResourceCategory.DATA, "Collection", "0").getObj();
    Map<Long, MetadataEntityInfoItemPO> infoItemMap = (Map<Long, MetadataEntityInfoItemPO>)this.metadataField.metadataFieldToInfoItem(metadataFieldWithElementVOS).stream().collect(Collectors.toMap(MetadataEntityInfoItemPO::getInfoItemId, Function.identity(), (k1, k2) -> k1));
    boolean hasIsDelete = false;
    boolean hasSyncStatus = false;
    boolean hasStartTime = false;
    boolean hasLastModifyTime = false;
    List<FieldMapping> fieldMappingList = new ArrayList<>(syncTaskFieldsList.size() + 4);
    for (DataSyncTaskFieldsPO po : syncTaskFieldsList) {
      if (po.getTargetFieldId() == null)
        continue; 
      FieldMapping fieldMapping = new FieldMapping();
      fieldMapping.setSource(po.getSourceFieldName());
      MetadataEntityInfoItemPO targetInfoItem = infoItemMap.get(po.getTargetFieldId());
      if (targetInfoItem != null) {
        fieldMapping.setTarget(targetInfoItem.getDataItemName());
        fieldMapping.setTargetPrimaryKey(Boolean.valueOf(DCConst.TRUE.equals(targetInfoItem.getIsPrimaryKey())));
      } 
      fieldMappingList.add(fieldMapping);
      if (fieldMapping.getTarget().equalsIgnoreCase(setting.getIsDelete())) {
        hasIsDelete = true;
        continue;
      } 
      if (fieldMapping.getTarget().equalsIgnoreCase(setting.getSyncStatus())) {
        hasSyncStatus = true;
        continue;
      } 
      if (fieldMapping.getTarget().equalsIgnoreCase(setting.getStartTime())) {
        hasStartTime = true;
        continue;
      } 
      if (fieldMapping.getTarget().equalsIgnoreCase(setting.getLastModifyTime()))
        hasLastModifyTime = true; 
    } 
    if (!hasIsDelete && setting.getIsDelete() != null)
      fieldMappingList.add(new FieldMapping(null, setting.getIsDelete(), Boolean.valueOf(false))); 
    if (!hasStartTime && setting.getStartTime() != null)
      fieldMappingList.add(new FieldMapping(null, setting.getStartTime(), Boolean.valueOf(false))); 
    if (!hasLastModifyTime && setting.getLastModifyTime() != null)
      fieldMappingList.add(new FieldMapping(null, setting.getLastModifyTime(), Boolean.valueOf(false))); 
    if (!hasSyncStatus && setting.getSyncStatus() != null)
      fieldMappingList.add(new FieldMapping(null, setting.getSyncStatus(), Boolean.valueOf(false))); 
    return fieldMappingList;
  }
  
  public List<ColumnStruct> getSourceColumns(List<DataSyncTaskFieldsPO> fields) {
    return (List<ColumnStruct>)fields.stream().filter(item -> (item.getSourceFieldName() != null))
      .map(item -> {
          ColumnStruct column = new ColumnStruct();
          column.setName(item.getSourceFieldName());
          String sourceFieldType = item.getSourceFieldType();
          String[] fieldTypeParts = sourceFieldType.split("\\(|\\)");
          if (fieldTypeParts.length > 0)
            column.setTypeName(fieldTypeParts[0].trim()); 
          if (fieldTypeParts.length > 1) {
            String[] fieldSizeParts = fieldTypeParts[1].split(",");
            String lengthPart = fieldSizeParts[0].trim();
            try {
              if (!"null".equals(lengthPart))
                column.setLength(Integer.valueOf(lengthPart)); 
              if (fieldSizeParts.length > 1)
                column.setScale(Integer.valueOf(fieldSizeParts[1].trim())); 
            } catch (NumberFormatException numberFormatException) {}
          } 
          return column;
        }).collect(Collectors.toList());
  }
  
  public DataSourceVO getNotEncryptDataSource(Long dataSourceId, String platCode, String tokenId) {
    DataSourceVO dataSource = (DataSourceVO)this.standardClient.getDataSourceInfoDoNotDecryptById(dataSourceId, platCode, tokenId).getObj();
    String decryptPassword = EncryptManagerFactory.getEncrypt("SM2").decrypt(dataSource.getPassword(), CoreConst.GroupKeyType.DATACENTER);
    dataSource.setPassword(decryptPassword);
    return dataSource;
  }
  
  private void checkIsTableExist(String tableName, ConnectInfo connectInfo) {
    DataSourceDTO dataSource = parseToDataSourceDTO(connectInfo);
    ResultEntity<DataStructProvider> result = DataStructProvider.getDataBaseProvider(dataSource, false, false);
    if (!DCConst.FALSE.equals(result.getCode()))
      throw new BusinessException(result.getCode(), result.getMsg()); 
    DataStructProvider provider = (DataStructProvider)result.getObj();
    if (!Boolean.TRUE.equals(provider.isExistsTable(tableName)))
      throw new BusinessException(Integer.valueOf(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_TABLE_NOT_EXIST.getCode()), 
          String.format(StatusCodeEnum.DATA_RESOURCE_DATA_SYNC_TABLE_NOT_EXIST.getMessage(), new Object[] { DatabaseHelper.getFullTableName(connectInfo.getDatabase(), connectInfo.getSchema(), tableName) })); 
  }
  
  private DataSourceDTO parseToDataSourceDTO(ConnectInfo connectInfo) {
    DataSourceDTO dataSource = new DataSourceDTO();
    dataSource.setType(connectInfo.getType());
    dataSource.setIpName(connectInfo.getHost());
    dataSource.setPort(connectInfo.getPort());
    dataSource.setUser(connectInfo.getUsername());
    dataSource.setPassword(connectInfo.getPassword());
    dataSource.setDbName(connectInfo.getDatabase());
    dataSource.setDbSchema(connectInfo.getSchema());
    return dataSource;
  }
  
  public Boolean createTargetTable(DataSyncTaskPO task, MetadataEntityVO entity, List<TableColumnsVO> targetColumns, DataStructProvider provider, String platCode, String tokenId) {
    Map<String, MetadataEntityInfoItemPO> infoItemMap = (Map<String, MetadataEntityInfoItemPO>)MetadataManager.getEntityInfoItemsByEntityId(entity.getEntityId(), DCConst.DataResourceCategory.DATA, platCode, tokenId).stream().collect(Collectors.toMap(MetadataEntityInfoItemPO::getDataItemName, Function.identity(), (k1, k2) -> k2));
    List<Long> dataItemIds = (List<Long>)infoItemMap.values().stream().map(MetadataEntityInfoItemPO::getDataItemId).distinct().collect(Collectors.toList());
    if (dataItemIds.isEmpty())
      return Boolean.valueOf(false); 
    Map<Long, MetadataEntityDataItemVO> dataItemMap = (Map<Long, MetadataEntityDataItemVO>)((List)this.standardClient.getMetadataEntityDataItemByIds(dataItemIds, "Collection", "0").getObj()).stream().collect(Collectors.toMap(MetadataEntityDataItemVO::getDataItemId, Function.identity(), (k1, k2) -> k2));
    Map<String, MetadataTypeVO> dataTypeMap = (Map<String, MetadataTypeVO>)((List)this.standardClient.getMetadataTypeList("Collection", "0").getObj()).stream().collect(Collectors.toMap(MetadataTypeVO::getDataTypeName, Function.identity(), (k1, k2) -> k2));
    List<ColumnParamDTO> columns = new ArrayList<>();
    boolean needIsDelete = Objects.equals(task.getDeleteMode(), Integer.valueOf(1));
    boolean needTimestamp = Objects.equals(task.getSyncMarkTimestamp(), Integer.valueOf(1));
    String isDelete = SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.IS_DELETE);
    String lastModifyTime = SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.LAST_MODIFY_TIME);
    for (TableColumnsVO targetColumn : targetColumns) {
      MetadataEntityInfoItemPO infoItem = infoItemMap.get(targetColumn.getFieldName());
      if (infoItem == null)
        continue; 
      MetadataEntityDataItemVO dataItem = dataItemMap.get(infoItem.getDataItemId());
      if (dataItem == null)
        continue; 
      ColumnParamDTO columnParam = assembleColumnParam(targetColumn, infoItem, dataItem, dataTypeMap, provider.getDbType());
      if (needIsDelete && isDelete.equals(targetColumn.getFieldName()))
        needIsDelete = false; 
      if (needTimestamp && lastModifyTime.equals(targetColumn.getFieldName()))
        needTimestamp = false; 
      columns.add(columnParam);
    } 
    Integer sysTimeLength = ("mysql".equals(provider.getDbType()) || "sqlserver".equals(provider.getDbType())) ? Integer.valueOf(3) : null;
    MetadataTypeVO sysTimeType = dataTypeMap.get("DateTime");
    if ("mysql".equals(provider.getDbType())) {
      sysTimeType.setMysqlDbType("datetime");
    } else if ("sqlserver".equals(provider.getDbType())) {
      sysTimeType.setSqlDbType("datetime2");
    } 
    columns.add(new ColumnParamDTO(SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.START_TIME), sysTimeType, sysTimeLength, Integer.valueOf(0), DCConst.TRUE, DCConst.FALSE));
    columns.add(new ColumnParamDTO(SysFieldHelper.getBaseSysFieldName(BaseSysFieldEnum.SYNC_STATUS), dataTypeMap.get("Int"), Integer.valueOf(0), Integer.valueOf(0), DCConst.TRUE, DCConst.FALSE));
    if (needIsDelete)
      columns.add(new ColumnParamDTO(isDelete, dataTypeMap.get("Int"), Integer.valueOf(0), Integer.valueOf(0), DCConst.TRUE, DCConst.FALSE)); 
    if (needTimestamp)
      columns.add(new ColumnParamDTO(lastModifyTime, sysTimeType, sysTimeLength, Integer.valueOf(0), DCConst.TRUE, DCConst.FALSE)); 
    TableParamDTO result = new TableParamDTO();
    result.setTableName(entity.getEntityTableName());
    if (!"mysql".equals(provider.getDbType()))
      result.setSchemaName(entity.getDataSchemaName()); 
    result.setColumns(columns);
    return provider.createTable(result);
  }
  
  private ColumnParamDTO assembleColumnParam(TableColumnsVO targetColumn, MetadataEntityInfoItemPO infoItem, MetadataEntityDataItemVO dataItem, Map<String, MetadataTypeVO> dataTypeMap, String dbType) {
    ColumnParamDTO columnParam = new ColumnParamDTO();
    columnParam.setColumnName(targetColumn.getFieldName());
    MetadataTypeVO standardType = dataTypeMap.get(dataItem.getDataType());
    Map<String, Object> typeInfo = FieldTypeValidatorHandler.parseFieldType(targetColumn.getFieldType());
    if (standardType == null)
      standardType = new MetadataTypeVO(); 
    MetadataTypeVO newType = new MetadataTypeVO();
    BeanUtils.copyProperties(standardType, newType);
    String type = (String)typeInfo.get("type");
    if ("mysql".equals(dbType)) {
      newType.setMysqlDbType(type);
    } else if ("sqlserver".equals(dbType)) {
      newType.setSqlDbType(type);
    } else if ("postgresql".equals(dbType)) {
      newType.setPgsqlDbType(type);
    } 
    columnParam.setColumnType(newType);
    columnParam.setColumnLength((targetColumn.getFieldSize() == null) ? (Integer)typeInfo.get("length") : targetColumn.getFieldSize());
    columnParam.setColumnScale((Integer)typeInfo.get("dataScale"));
    if (DCConst.TRUE.equals(infoItem.getIsPrimaryKey())) {
      columnParam.setIsNull(DCConst.FALSE);
      columnParam.setIsPrimaryKey(DCConst.TRUE);
    } else {
      columnParam.setIsNull(DCConst.TRUE);
      columnParam.setIsPrimaryKey(DCConst.FALSE);
    } 
    return columnParam;
  }
  
  private boolean notSuppotDatabaseType(String type) {
    return (!"mysql".equals(type) && !"sqlserver".equals(type) && !"postgresql".equals(type));
  }
}
